/*
 * Analyze the byte distribution of stdin.
 *
 * Dino Dai Zovi <ghandi@mindless.com>
 */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#define READ_SIZE 1024

int main(int argc, char* argv[])
{
    int f, n, i, total = 0, ascii = 0, print = 0;
    unsigned char b[READ_SIZE], x = 0;
    char c, p = 0;
    unsigned int bytes[256];

    while ((c = getopt(argc, argv, "x:p")) != EOF) {
        switch (c) {
        case 'x':
            /* XOR w/ specified value */
            x = strtoul(optarg, NULL, 16);
            break;
        case 'p':
            p = 1;
        default:
            printf("usage: %s -x value -p\n", argv[0]);
            break;
        }
    }
    
    memset(bytes, 0, 256 * sizeof(unsigned int));
    
    while ((n = read(0, &b, READ_SIZE)) > 0)
        for (i = 0; i < n; i++) {
            unsigned char b2 = x ^ b[i];
            bytes[b2]++;
            total++;
            if (isascii(b2)) ascii++;
            if (isprint(b2)) print++;
        }

    if (p) {
        for (i = 0; i < 256; i++) {
            printf("0x%.2x (", i);
            if (isprint(i))
                printf("%c", (char)i);
            else
                printf(".", i);
            printf("): %d\n", bytes[i]);
        }
    }
        
    printf("%.2f%% ASCII (%.2f%% printable)\n",
           100 * ((float)ascii/(float)total),
           100 * ((float)print/(float)total));
    
    return 0;
}
